@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1998-2003 Apple Computer, Inc. All rights reserved.
+ * Copyright (c) 1998-2005 Apple Computer, Inc. All rights reserved.
  *
  * @APPLE_LICENSE_HEADER_START@
  * 
@@ -51,10 +51,6 @@ OSDefineMetaClassAndStructors(IOApplePartitionScheme, IOPartitionScheme);
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 
-#define kIOApplePartitionSchemeContentTable "Content Table"
-
-// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-
 bool IOApplePartitionScheme::init(OSDictionary * properties)
 {
     //
@@ -287,7 +283,7 @@ OSSet * IOApplePartitionScheme::scan(SInt32 * score)
 
         // Determine whether the partition entry signature is present.
 
-        if (OSSwapBigToHostInt16(dpmeMap->dpme_signature) == DPME_SIGNATURE)
+        if ( OSSwapBigToHostInt16(dpmeMap->dpme_signature) == DPME_SIGNATURE )
         {
             dpmeBlockSize = sizeof(dpme);             // (old school block size)
             dpmeOldSchool = true;
@@ -496,24 +492,6 @@ IOMedia * IOApplePartitionScheme::instantiateMediaObject(
         partitionSize = media->getSize() - partitionBase;
     }
 
-    // Look up a type for the new partition.
-
-    OSDictionary * hintTable = OSDynamicCast( 
-              /* type     */ OSDictionary,
-              /* instance */ getProperty(kIOApplePartitionSchemeContentTable) );
-
-    if ( hintTable )
-    {
-        OSString * hintValue = OSDynamicCast( 
-                       /* type     */ OSString,
-                       /* instance */ hintTable->getObject(partitionHint) );
-
-        if ( hintValue )
-        {
-            strncmp(partitionHint, hintValue->getCStringNoCopy(), DPISTRLEN);
-        }
-    }
-
     // Determine whether the new partition type is Apple_Free, which we choose
     // not to publish because it is an internal concept to the partition map.
 
